#include "MemoryFunctions.h"

HRESULT ReadVirtual(IDebugClient* pDebugClient, ULONG64 Offset, PVOID Buffer, ULONG BufferSize, PULONG BytesRead)
{
    HRESULT hr = E_FAIL;

    // Reset Buffer
    memset(Buffer, 0, BufferSize);

    IDebugDataSpaces* pDebugDataSpaces;
    if (SUCCEEDED(hr = pDebugClient->QueryInterface(__uuidof(IDebugDataSpaces), (void **)&pDebugDataSpaces)))
    {
        hr = pDebugDataSpaces->ReadVirtual(SignExtend(pDebugClient, Offset), Buffer, BufferSize, BytesRead);
        pDebugDataSpaces->Release();
    }
    return hr;
}

HRESULT ReadVirtualC(IDebugClient* pDebugClient, ULONG64 Offset, PVOID left, PVOID right, ULONG BufferSize, PULONG BytesRead)
{
	HRESULT hr = E_FAIL;

    // Reset Buffer
    memset(left, 0, BufferSize);
	memset(right, 0, BufferSize);

	PVOID *bufferC = new PVOID;

    IDebugDataSpaces* pDebugDataSpaces;
    if (SUCCEEDED(hr = pDebugClient->QueryInterface(__uuidof(IDebugDataSpaces), (void **)&pDebugDataSpaces)))
    {
       // hr = pDebugDataSpaces->ReadVirtual(SignExtendF(pDebugClient, Offset), bufferC, BufferSize, BytesRead);

		////Read Children:
		//ULONG64 children = (ULONG64) *bufferC;
		hr = pDebugDataSpaces->ReadVirtual(SignExtend(pDebugClient, Offset), left, BufferSize, BytesRead);
		hr = pDebugDataSpaces->ReadVirtual(SignExtend(pDebugClient, Offset + 8), right, BufferSize, BytesRead);
        pDebugDataSpaces->Release();
    }
    return hr;
}



ULONG64 SignExtend(IDebugClient* pDebugClient, ULONG64 ptr)
{
    if ((ptr > 0x7FFFFFFF) && (ptr <= 0xFFFFFFFF))
    {
        IDebugControl* pDebugControl;
        if (SUCCEEDED(pDebugClient->QueryInterface(__uuidof(IDebugControl), (void **)&pDebugControl)))
        {
            if (pDebugControl->IsPointer64Bit() == S_FALSE)
            {
                ptr += 0xFFFFFFFF00000000;
            }
            pDebugControl->Release();
        }
    }
    return ptr;
}

HRESULT ReadPage(IDebugClient* pDebugClient, ULONG pfnNo, PVOID Buffer)
{
	HRESULT hRes = E_FAIL;
	IDebugDataSpaces *pPhysicalSpace;
	ULONG PAGE_SIZE = 4096;
	if (SUCCEEDED ( hRes = pDebugClient->QueryInterface(__uuidof(IDebugDataSpaces), (void **)&pPhysicalSpace)))
	{
		ULONG64 pfn = (ULONG64) pfnNo;
		ULONG64 address = pfn * 0x1000;
		ULONG *puSize = new ULONG;
		hRes = pPhysicalSpace->ReadPhysical(address, Buffer, PAGE_SIZE, puSize);
		if(hRes != S_OK)
		{
			ULONG readBytes = *puSize;
		}
		pPhysicalSpace->Release();
		delete[] puSize;
	}
	return hRes;
}

HRESULT SearchNonPaged(IDebugClient *pDebugClient, PVOID pattern, ULONG patSize, ULONG64 length, PULONG64 result )
{
	HRESULT hRes = E_FAIL;
	IDebugDataSpaces *pNonPagedSpace;
	if (SUCCEEDED( hRes = pDebugClient->QueryInterface(__uuidof(IDebugDataSpaces), (void **)&pNonPagedSpace)))
	{
		ULONG64 nonPagedOffset = 0xffffe00000000000;
		hRes = pNonPagedSpace->SearchVirtual(nonPagedOffset, length, pattern, patSize, 1, result);
		pNonPagedSpace->Release();

	}

	return hRes;
}