//ProcessFunctions.cpp
//By Hans Kristian Brendmo

#include "ProcessFunctions.h"

HRESULT ReadLockedPagesList(IDebugClient* pDebugClient, PCSTR args, PVOID output)
{
	HRESULT hr = E_FAIL;
	DEBUG_VALUE Value = {0};
	ULONG *pOffset = new ULONG;
	IDebugControl *pDebugControl;

	hr = FindFieldOffset(pDebugClient,"LockedPagesList", "_EPROCESS", "ntdll", pOffset);

	if (SUCCEEDED( hr = pDebugClient->QueryInterface(__uuidof(IDebugControl), (void **)&pDebugControl))) 
	{
		if( SUCCEEDED ( hr = pDebugControl->Evaluate(args,DEBUG_VALUE_INT64,&Value, NULL)))
		{
			ULONG64 ulLockPages = Value.I64 + *pOffset;
			ULONG uSize = 8;
			if( SUCCEEDED ( hr = ReadVirtual(pDebugClient, ulLockPages, output, uSize, NULL)))
			{
				if(output == NULL)
				{
					pDebugControl->Output(DEBUG_OUTPUT_ERROR, "LockedPagesList could not be read, perhaps lock tracking is disabled.");
					output = 0;
				}
				else
				{
					//pDebugControl->Output(DEBUG_OUTPUT_NORMAL, "Read addr: %p  \n",*pBuf);
				}
			}

		}
		pDebugControl->Release();
	}

	delete[] pOffset;
	return hr;
}

HRESULT ReadProcessName(IDebugClient* pDebugClient, ULONG64 eProcess, PVOID nameBuf)
{
	HRESULT hr = E_FAIL;
	ULONG *pOffset = new ULONG;
	ULONG nameChars = 15;

	if( SUCCEEDED (hr = FindFieldOffset(pDebugClient,"ImageFileName", "_EPROCESS", "ntdll", pOffset)))
	{
		ULONG64 nameOffset = eProcess + (ULONG64) *pOffset;
		if( SUCCEEDED (hr = ReadVirtual(pDebugClient, nameOffset, nameBuf, nameChars, NULL) ) )
		{

		}
		else
		{
			OutputError(pDebugClient, "Error when reading virtual address. HRESULT %u \n",hr);
		}
	}
	else
	{
		OutputError(pDebugClient, "Error when reading field offset. HRESULT %u \n",hr);
	}

	delete[] pOffset;	
	return hr;
}

HRESULT ReadLockedPagesList(IDebugClient* pDebugClient, ULONG64 arg, PVOID output)
{
	HRESULT hr = E_FAIL;
	DEBUG_VALUE Value = {0};
	ULONG *pOffset = new ULONG;
	IDebugControl *pDebugControl;

	hr = FindFieldOffset(pDebugClient,"LockedPagesList", "_EPROCESS", "ntdll", pOffset);

	if (SUCCEEDED( hr = pDebugClient->QueryInterface(__uuidof(IDebugControl), (void **)&pDebugControl))) 
	{
		ULONG uSize = 8;
		if( SUCCEEDED ( hr = ReadVirtual(pDebugClient, arg+*pOffset, output, uSize, NULL)))
		{
			if(output == NULL)
			{
				pDebugControl->Output(DEBUG_OUTPUT_ERROR, "LockedPagesList could not be read, perhaps lock tracking is disabled.");
				output = 0;
			}
			else
			{
				//pDebugControl->Output(DEBUG_OUTPUT_NORMAL, "Read addr: %p  \n",*pBuf);
			}
		}
		pDebugControl->Release();
	}
	delete[] pOffset;
	return hr;
}



HRESULT GetCurrentProcess(IDebugClient *pDebugClient, PULONG64 process)
{
	HRESULT hRes = E_FAIL;
	IDebugSystemObjects* pSystemObjects;

	if(SUCCEEDED (hRes = pDebugClient->QueryInterface(__uuidof(IDebugSystemObjects), (void **)&pSystemObjects)))
	{
			pSystemObjects->GetCurrentProcessDataOffset(process);
			pSystemObjects->Release();
	}
	return hRes;
}

HRESULT GetNextProcess(IDebugClient *pDebugClient, PULONG64 currProcess, PULONG64 nextProcess)
{
	HRESULT hRes = E_FAIL;
	ULONG *pLink = new ULONG;
	ULONG pointerSize = 8;

	if( SUCCEEDED (hRes = FindFieldOffset(pDebugClient,"ActiveProcessLinks","_EPROCESS", "ntdll", pLink)))
	{
		ULONG64 nextOffset = *currProcess + (ULONG64) *pLink;
		PVOID *buf = new PVOID;
		if (SUCCEEDED (hRes = ReadVirtual(pDebugClient,  nextOffset, buf, pointerSize, NULL)))
		{
			*nextProcess = (ULONG64) *buf;
		}
		else
		{
			*nextProcess = 0;
		}
		delete[] buf;
	}

	delete[] pLink;
	return hRes;
}

HRESULT TraverseLockedPages(IDebugClient* pDebugClient)
{
	HRESULT hRes = E_FAIL;
	ULONG64* currAdr = new ULONG64;
	ULONG64 *nextAdr  = new ULONG64;
	ULONG64 startAdr = 0;
	IDebugControl *pDebugControl;
	if( SUCCEEDED (hRes = GetCurrentProcess(pDebugClient, currAdr)))
	{
		//Traverse all process:
		startAdr = *currAdr;
		PROCESS *currProcess = new PROCESS(pDebugClient, currAdr);
		if(SUCCEEDED (hRes = pDebugClient->QueryInterface(__uuidof(IDebugControl), (void **)&pDebugControl)))
		{
			*nextAdr = currProcess->getNextProcess(pDebugClient);
			while ( startAdr != *nextAdr  )
			{	
				
				if(currProcess->hasLockedPages())
				{
					pDebugControl->Output
					(
						DEBUG_OUTPUT_NORMAL, 
						"Process: %p [%s] ,locked pages: %u \n", 
						currProcess->getEprocess(),
						currProcess->getName(),
						currProcess->getCount()
					);

				}

				currProcess->update(pDebugClient, nextAdr);
				*nextAdr = currProcess->getNextProcess(pDebugClient);
				
				while( !currProcess->isValid() )
				{
					currProcess->update(pDebugClient, nextAdr);
					*nextAdr = currProcess->getNextProcess(pDebugClient);
				}

			}
			pDebugControl->Release();
		}
		else
		{
			OutputError(pDebugClient, "Failed to query interface. HRESULT %u", hRes);
		}
	}
	else
	{
		OutputError(pDebugClient, "Failed to get current process. HRESULT %u", hRes);
	}
	delete[] nextAdr;
	delete[] currAdr;
	return hRes;	
}