
#include "RetrieveMdl.h"

HRESULT FindValidMdls(IDebugClient *pDebugClient)
{
	HRESULT hRes = E_FAIL;
	PVOID *mdlPat = new PVOID;
	ULONG64 *result = new ULONG64;
	PULONG64 startAdr = new ULONG64;
	ULONG64 numBytesToSearch = 10000000000;
	*mdlPat = (PVOID) 0x206C644D;
	IDebugControl *pDebugControl;

	if (SUCCEEDED( hRes = pDebugClient->QueryInterface(__uuidof(IDebugControl), (void **)&pDebugControl))) 
	{
		//hRes = SearchNonPaged(pDebugClient, mdlPat, 4, numBytesToSearch, result);
		pDebugControl->Output(DEBUG_OUTPUT_NORMAL, "Finding starting address... \n" );
		if( SUCCEEDED (SearchForStart(pDebugClient, startAdr)))
		{
			pDebugControl->Output(DEBUG_OUTPUT_NORMAL, "Start address found at %p... \n", *startAdr);
			BOOL mdlFound = false;
			ULONG64 offset = 10000000;
			while(!mdlFound && *startAdr < (0xffffe00000000000 + numBytesToSearch))
			{
				hRes = SearchForMdlHeader(pDebugClient, *startAdr, offset, result);

				if(hRes == S_OK)
				{
					 if(ValidHeader(pDebugClient, *result))
					 {
						 //mdlFound = true;
						 ValidMdl(pDebugClient, *result+0x10);
						 *startAdr = *result + 8;
						 					
					 }
					 else
					 {
						 *startAdr = *result + 8;
					 }
				}
				else
				{
					pDebugControl->Output(DEBUG_OUTPUT_ERROR, "Pattern not found CurrAdr %p \n", *startAdr);
					*startAdr += offset;
				}
				if( pDebugControl->GetInterrupt() == S_OK)
				{
					pDebugControl->Output(DEBUG_OUTPUT_NORMAL, "Search was interrupted, Curradr: %p \n",*startAdr);
					break;
				}
			}
		}
		pDebugControl->Release();
	}
	delete[] mdlPat;
	delete[] result;
	delete[] startAdr;
	return hRes;
}


//Returns address at start of MDL pool header:
HRESULT SearchForMdlHeader(IDebugClient *pDebugClient, ULONG64 startAdr, ULONG64 length,  PULONG64 result)
{
	HRESULT hRes = E_FAIL;
	ULONG64 currAdr = startAdr;
	ULONG64 endAdr = startAdr + length;
	BOOL notFound = true;	
	UINT pattern = 0x206C644D;
	IDebugControl *pDebugControl;
	const UINT pageSize = 4096;
	PVOID *buf = new PVOID;

	if (SUCCEEDED( hRes = pDebugClient->QueryInterface(__uuidof(IDebugControl), (void **)&pDebugControl))) 
	{
		
		while( currAdr < endAdr && notFound)
		{
			
			if( SUCCEEDED (hRes = ReadVirtual(pDebugClient, currAdr, buf, 4, NULL)))
			{
				if( (UINT) *buf == pattern)
				{
					notFound = false;
					*result = currAdr - 4;
				}
				
			}
			currAdr += 4;
			

			/*
			if( SUCCEEDED (hRes = ReadVirtual(pDebugClient, currAdr, buf, pageSize, NULL)))
			{
				UINT pos[pageSize];
				UINT numMatches = SearchBuffer(buf, pageSize, pos);
				for(UINT i = 0; i < numMatches; i++)
				{

				}
				
			}
			currAdr += 4;*/
		
		}
		pDebugControl->Release();
	}
	if(notFound)
	{
		hRes = E_FAIL;
	}
	delete[] buf;
	return hRes;

}

//Find a viable starting address for the nonpaged pool:
HRESULT SearchForStart(IDebugClient *pDebugClient, PULONG64 result)
{
	IDebugControl *pDebugControl;
	HRESULT hRes = E_FAIL;
	if (SUCCEEDED( hRes = pDebugClient->QueryInterface(__uuidof(IDebugControl), (void **)&pDebugControl))) 
	{
		ULONG64 currAdr = 0xffffe00000000000;
		BOOL foundStart = false;
		const ULONG bufSize = 4096;
		PVOID buf[bufSize];
		while(!foundStart)
		{
			if(SUCCEEDED(hRes = ReadVirtual(pDebugClient, currAdr, buf, bufSize, NULL)))
			{
				foundStart = true;
				*result = currAdr;
			}
			if( pDebugControl->GetInterrupt() == S_OK)
			{
				pDebugControl->Output(DEBUG_OUTPUT_NORMAL, "Search was interrupted, Curradr: %p \n",currAdr);
				break;
			}
			currAdr += bufSize;
		}
	}

	return hRes;
}

BOOL ValidHeader(IDebugClient *pDebugClient, ULONG64 headerAdr)
{
	BOOL valid = false;
	POOL *mdlH = new POOL(pDebugClient, headerAdr);
	IDebugControl *pDebugControl;
	if (SUCCEEDED(pDebugClient->QueryInterface(__uuidof(IDebugControl), (void **)&pDebugControl))) 
	{
		if( mdlH->getValid() )
		{
			//pDebugControl->Output(DEBUG_OUTPUT_NORMAL, "Found valid mdl pool header at:%p \n", mdlH->getAdr());
			valid = true;
		}
		else
		{
			//pDebugControl->Output(DEBUG_OUTPUT_NORMAL, "No valid mdl pool header found. \n");
		}
		pDebugControl->Release();
	}
	delete mdlH;
	return valid;
}

BOOL ValidMdl(IDebugClient *pDebugClient, ULONG64 mdlAdr)
{
	BOOL success = false;
	MDL *mdl = new MDL(pDebugClient, mdlAdr);
	IDebugControl *pDebugControl;
	if (SUCCEEDED(pDebugClient->QueryInterface(__uuidof(IDebugControl), (void **)&pDebugControl))) 
	{
		if(mdl->getProc() == 0xffffe001b547e840)
		{
			pDebugControl->Output(DEBUG_OUTPUT_NORMAL, "\nByte count: %u \n", mdl->getByteCount());
			pDebugControl->Output(DEBUG_OUTPUT_NORMAL, "Byte offset: %u \n", mdl->getByteOffset());
			pDebugControl->Output(DEBUG_OUTPUT_NORMAL, "Process: %p \n", mdl->getProc());
			pDebugControl->Output(DEBUG_OUTPUT_NORMAL, "Mdl address: %p \n", mdlAdr);
			pDebugControl->Output(DEBUG_OUTPUT_NORMAL, "Linked: %p \n \n", mdl->getLinked());
			pDebugControl->Release();
		}
	}
	return success;
}

UINT SearchBuffer(PVOID *buf, UINT bufSize, UINT *positions)
{
	BYTE pat1 = 0x20;
	BYTE pat2 = 0x6C;
	BYTE pat3 = 0x64;
	BYTE pat4 = 0x4D;
	UINT patSize = 4;
	UINT timesFound = 0;

	for(UINT i = 0; i < bufSize - patSize;)
	{
		if( (BYTE) buf[i] != pat1)
		{
			i += patSize;
		}
		else if ( (BYTE) buf[i+1] != pat2)
		{
			i+= 3;
		}
		else if ( (BYTE) buf[i+2] != pat3)
		{
			i+= 2;
		}
		else if ( (BYTE) buf[i+3] == pat4)
		{
			i += 1;
			positions[timesFound] = i - patSize;
			timesFound++;
		}
		else
		{
			i+=1;
		}
	}
	return timesFound;
}