//Supportfunctions.cpp

#include "SupportFunctions.h"
#include <ctime>
#include "CEventCallbacks.h"
#include "NODE.h"
#include "POOL.h"
#include "MemoryFunctions.h"
#include "PFNdata.h"

VOID CALLBACK TimeoutCallback (PTP_CALLBACK_INSTANCE pInstance, PVOID pvContext, PTP_TIMER pTimer)
{
	UNREFERENCED_PARAMETER(pInstance);
    UNREFERENCED_PARAMETER(pvContext);
    UNREFERENCED_PARAMETER(pTimer);
}

HRESULT SetBreakpoints(IDebugClient* pDebugClient)
{
	HRESULT hRes = E_FAIL;
	IDebugControl* pDebugControl;
	if (SUCCEEDED(hRes = pDebugClient->QueryInterface(__uuidof(IDebugControl), (void **)&pDebugControl)))
	{
		hRes = pDebugControl->Execute(DEBUG_OUTCTL_ALL_CLIENTS,"bp nt!MmProbeAndLockPages",DEBUG_EXECUTE_NO_REPEAT);
		//hRes = pDebugControl->Execute(DEBUG_OUTCTL_ALL_CLIENTS,"bp ntdll!NtCreateFIle",DEBUG_EXECUTE_NO_REPEAT);
		pDebugControl->Release();
	}	
	return hRes;
}

HRESULT ClearBreakpoints(IDebugClient* pDebugClient)
{
	HRESULT hRes = E_FAIL;
	IDebugControl* pDebugControl;
	if (SUCCEEDED(hRes = pDebugClient->QueryInterface(__uuidof(IDebugControl), (void **)&pDebugControl)))
	{
		ULONG* status = new ULONG;
		pDebugControl->GetExecutionStatus(status);
		if(*status == DEBUG_STATUS_BREAK)
		{
			hRes = pDebugControl->Execute(DEBUG_OUTCTL_ALL_CLIENTS,"bc *",DEBUG_EXECUTE_NO_REPEAT);
			pDebugControl->Output(DEBUG_OUTPUT_NORMAL,"All breakpoints cleared \n");
		}
		else if(*status == DEBUG_STATUS_GO)
		{
			hRes = pDebugControl->SetInterrupt(DEBUG_INTERRUPT_ACTIVE);
			if(hRes == S_OK)
			{
				hRes = pDebugControl->Execute(DEBUG_OUTCTL_ALL_CLIENTS,"bc *",DEBUG_EXECUTE_NO_REPEAT);
			}
			else
			{
				pDebugControl->Output(DEBUG_OUTPUT_ERROR, "Break-in failed. HRESULT: %x \n",hRes);
			}
		}
		else
		{
			hRes = E_FAIL;
		}
		delete[] status;
		pDebugControl->Release();
	}	
	return hRes;
}


HRESULT TraverseTree(IDebugClient* pDebugClient, PVOID* input)
{

	//Make root node:
	NODE *rootNode = new NODE( (ULONG64) *input, pDebugClient);
	IDebugControl* pDebugControl;
	
    if (SUCCEEDED(pDebugClient->QueryInterface(__uuidof(IDebugControl), (void **)&pDebugControl)))
	{
		//In-order tree traversal:
		std::stack<ULONG64> s;
		NODE *currNode = rootNode;
		s.push(rootNode->adr() );
		UINT totalsize = 0;
		while( !s.empty() )
		{

			currNode->setadr(s.top(), pDebugClient);
			s.pop();

			pDebugControl->Output(DEBUG_OUTPUT_NORMAL, "Lock tracker: %p, Length %u bytes \n", currNode->adr(), currNode->length() );
			MDL *md = new MDL();
			if(SUCCEEDED (MdlData(pDebugClient, currNode->adr(), md)))
			{
				pDebugControl->Output(DEBUG_OUTPUT_NORMAL, "	Mdl: %p, Owning process: %p, Linked: %u \n \n", md->getAddr(), md->getProc(), md->getLinked());
			}
			delete md;
			totalsize += currNode->length();

			if( currNode->right() > 0 )
			{
				s.push(currNode->right() );
			}
			if( currNode->left() > 0 )
			{
				s.push(currNode->left() );
			}

		}
		pDebugControl->Output(DEBUG_OUTPUT_NORMAL, "Total size: %u bytes \n", totalsize );
		delete currNode;
		pDebugControl->Release();
	}
	return S_OK;
}

HRESULT StartLogging(IDebugClient* pDebugClient)
{
	IDebugControl* pDebugControl;
	if (SUCCEEDED(pDebugClient->QueryInterface(__uuidof(IDebugControl), (void **)&pDebugControl)))
	{
		ULONG* status = new ULONG;
		pDebugControl->GetExecutionStatus(status);
		if(*status == DEBUG_STATUS_BREAK)
		{		
			if( SUCCEEDED (pDebugControl->SetExecutionStatus(DEBUG_STATUS_GO)))
			{
				pDebugControl->Output(DEBUG_OUTPUT_NORMAL, "Continuing execution... \n");

				//register event callback for client:
				//CEventCallbacks* pEventCallback = new CEventCallbacks;
				//pDebugClient->SetEventCallbacks(pEventCallback);

				//Perform logging for x seconds:
				time_t start;
				time_t end;
				double difference = 0;
				double seconds = 3.0;

				time(&start);

				while(difference < seconds)
				{
					//do work
					
					pDebugControl->WaitForEvent(0, 1000);
					pDebugControl->GetExecutionStatus(status);
					pDebugControl->Output(DEBUG_OUTPUT_NORMAL, "Execution status: %u \n  ", *status);
					time(&end);
					difference = difftime(end,start);
				}

			}
			else
			{
				pDebugControl->Output(DEBUG_OUTPUT_ERROR, "Failed to change execution status. \n");
				return E_FAIL;
			}
		}
		delete[] status;
		pDebugControl->Release();
	}
	else
	{
		return E_FAIL;
	}
	return S_OK;
}



HRESULT OutputError(IDebugClient* pDebugClient,  PCSTR message, HRESULT err)
{
	IDebugControl *pDebugControl;
	HRESULT hRes = E_FAIL;
	if (SUCCEEDED( hRes = pDebugClient->QueryInterface(__uuidof(IDebugControl), (void **)&pDebugControl)))
	{
		pDebugControl->Output(DEBUG_OUTPUT_ERROR, message, err);
		pDebugControl->Release();
	}
	return hRes;
}

HRESULT ReadPfnData(IDebugClient * pDebugClient, PCSTR szNum)
{
	IDebugControl *pDebugControl;
	HRESULT hRes = E_FAIL;
	if (SUCCEEDED( hRes = pDebugClient->QueryInterface(__uuidof(IDebugControl), (void **)&pDebugControl)))
	{
		ULONG num = 0;
		DEBUG_VALUE Value = {0};
		pDebugControl->Evaluate(szNum, DEBUG_VALUE_INT64, &Value, NULL);
		num = Value.I32;
		ULONG numLockedPages = 0;
		for(ULONG i = 1; i <= num; i++)
		{
			PFNdata *p = new PFNdata(pDebugClient, i);
			if( p->isValid() && p->isDriverLocked())
			{
				numLockedPages++;
			}
			delete p;
		}
		pDebugControl->Output(DEBUG_OUTPUT_NORMAL, "There are %u pages locked \n", numLockedPages);
		pDebugControl->Release();
	}
	return hRes;
}

HRESULT ReadPfnDatabase(IDebugClient * pDebugClient)
{
	IDebugControl *pDebugControl;
	HRESULT hRes = E_FAIL;
	if (SUCCEEDED( hRes = pDebugClient->QueryInterface(__uuidof(IDebugControl), (void **)&pDebugControl)))
	{
		ULONG64 currAddress = 0;
		ULONG numLockedPages = 0;
		ULONG pfns = 1;
		ULONG mult = 1;
		while( currAddress < 0xFFFFFA8008000000) 
		{
			PFNdata *p = new PFNdata(pDebugClient, pfns);
			if( p->isValid() && p->isDriverLocked())
			{
				if(p->getProc() == 0xffffe001b547e840)
				{
					numLockedPages++;
				}
			}
			pfns++;
			if( (pfns % 0x200000) == 0)
			{
				pDebugControl->Output(DEBUG_OUTPUT_NORMAL, "%u GB out of 512 GB scanned. \n", (mult*4));
				pDebugControl->Output(DEBUG_OUTPUT_NORMAL, "%u locked pages found. \n", numLockedPages);
				mult++;
			}
			if( pDebugControl->GetInterrupt() == S_OK)
			{
				pDebugControl->Output(DEBUG_OUTPUT_NORMAL, "The process was interrupted. \n");
				break;
			}
			currAddress = p->getAdr();
			delete p;
		}
		pDebugControl->Output(DEBUG_OUTPUT_NORMAL, "\nThere are %u pages locked. \n", numLockedPages);
		pDebugControl->Release();
	}
	return hRes;
}

void TestWrite()
{
	const ULONG BUFSIZE = 4096;
	LPCVOID *buf2 = new LPCVOID[BUFSIZE + 1];
	memset(buf2, '-', 4096);
	HANDLE hFile = CreateFile
					(
						TEXT("\\\\.\\D:\\Temp\\Myfile.dat"),
						GENERIC_WRITE,
						NULL,
						NULL,
						CREATE_NEW,
						FILE_ATTRIBUTE_NORMAL,
						NULL
					);
	WriteFile(hFile, buf2, 4096, NULL, NULL);
	delete[] buf2;
	CloseHandle(hFile);
}

HRESULT MdlData(IDebugClient *pDebugClient, ULONG64 adr, MDL *m)
{
	HRESULT hRes = E_FAIL;
	//Find mdl offset:
	ULONG *offset = new ULONG;
	if(SUCCEEDED (hRes = FindFieldOffset(pDebugClient, "Mdl", "_LOCK_TRACKER", "nt", offset)))
	{
		//Read mdl adr:
		ULONG64 mdlAddr = adr + (ULONG64) *offset;
		PVOID *buf = new PVOID;
		if( SUCCEEDED (hRes = ReadVirtual(pDebugClient, mdlAddr, buf, 8, NULL)))
		{
			m->updateAddr(pDebugClient, (ULONG64) *buf);
			if (!m->isValid())
			{
				hRes = E_FAIL;
			}

		}
	}
	return hRes;
	
}

HRESULT ReadPoolHeader(IDebugClient *pDebugClient, PCSTR arg)
{
	IDebugControl *pDebugControl;
	HRESULT hRes = E_FAIL;
	if (SUCCEEDED( hRes = pDebugClient->QueryInterface(__uuidof(IDebugControl), (void **)&pDebugControl)))
	{
		DEBUG_VALUE Value = {0};
		if(SUCCEEDED (hRes = pDebugControl->Evaluate(arg, DEBUG_VALUE_INT64, &Value, NULL)))
		{
			ULONG64 poolAdr = Value.I64;
			POOL *pH = new POOL(pDebugClient, poolAdr);
			pH->outputData(pDebugClient);
			delete pH;
		}
		pDebugControl->Release();
	}

	return hRes;
}


HRESULT WriteVM(IDebugClient * pDebugClient)
{
	IDebugControl *pDebugControl;
	HRESULT hRes = E_FAIL;
	//Open file to write:
	const ULONG BUFSIZE = 4096;
	PVOID buf[BUFSIZE];
	memset(buf, 0, BUFSIZE);
	FILE *fptr;
	fptr = NULL;
	fptr = fopen("d:\\Temp\\MyFile.dat","a");
	ULONG pagesWritten = 0;
	if (SUCCEEDED( hRes = pDebugClient->QueryInterface(__uuidof(IDebugControl), (void **)&pDebugControl)))
	{
		ULONG64 currAddress = 0;
		ULONG numLockedPages = 0;
		ULONG pfns = 1;
		ULONG mult = 1;
		while( currAddress < 0xFFFFFA8008000000) 
		{
			PFNdata *p = new PFNdata(pDebugClient, pfns);
			if( p->isValid() && p->isDriverLocked())
			{
				if(p->getProc() == 0xffffe001b547e840)
				{
					numLockedPages++;
					//writefile:
					if(fptr != NULL)
					{
						hRes = ReadPage(pDebugClient, pfns, buf);
						if(hRes == S_OK)
						{
							fwrite(buf, sizeof(BYTE), 4096, fptr);
							pagesWritten++;
						}
						else
						{
							pDebugControl->Output(DEBUG_OUTPUT_ERROR, "Failed to write page, pfn:%x , HRESULT: %x \n", pfns,hRes );
						}
					}
				}
			}
			pfns++;
			if( (pfns % 0x200000) == 0)
			{
				pDebugControl->Output(DEBUG_OUTPUT_NORMAL, "%u GB out of 512 GB scanned. \n", (mult*4));
				pDebugControl->Output(DEBUG_OUTPUT_NORMAL, "%u locked pages found. \n", numLockedPages);
				mult++;
			}
			if( pDebugControl->GetInterrupt() == S_OK)
			{
				pDebugControl->Output(DEBUG_OUTPUT_NORMAL, "The process was interrupted. \n");
				break;
			}
			currAddress = p->getAdr();
			delete p;
		}
		pDebugControl->Output(DEBUG_OUTPUT_NORMAL, "\nThere are %u pages locked. \n", numLockedPages);
		pDebugControl->Output(DEBUG_OUTPUT_NORMAL, "\n %u pages was written to disk \n", pagesWritten);
		pDebugControl->Release();
	}
	if(fptr != NULL)
	{
		fclose(fptr);
	}
	return hRes;
}

HRESULT WriteMem(IDebugClient *pDebugClient, ULONG64 start, UINT length, PCSTR fileName)
{
	HRESULT hRes = E_FAIL;
	HANDLE hFile = CreateFileA(
		fileName,
		GENERIC_WRITE,
		0,
		NULL,
		CREATE_NEW,
		FILE_ATTRIBUTE_NORMAL,
		NULL);
	IDebugControl *pDebugControl;
	const ULONG BUFSIZE = 4096;
	PVOID buf[BUFSIZE];
	memset(buf, 0, BUFSIZE);
	ULONG64 currAddress = start;
	ULONG BITMASK = 0xFFFFFFFF;
	UINT pagesWritten = 0;

	if(hFile != INVALID_HANDLE_VALUE)
	{
		if (SUCCEEDED( hRes = pDebugClient->QueryInterface(__uuidof(IDebugControl), (void **)&pDebugControl)))
		{
		
			for(UINT i = 0; i < length; i++)
			{
				PVOID *currPfn = new PVOID;

				if(SUCCEEDED (hRes = ReadVirtual(pDebugClient, currAddress, currPfn, 8, NULL)))
				{
					//Convert pfn to 32 bit no:
					ULONG pfn = (ULONG) *currPfn & BITMASK;
					if(SUCCEEDED (hRes = ReadPage(pDebugClient, pfn, buf)))
					{
						DWORD bytesWritten = 0;
						if(WriteFile(hFile, buf, BUFSIZE, &bytesWritten, NULL))
						{
							pagesWritten++;
						}
					}
				}
				currAddress += 0x8;
				delete[] currPfn;
			}
			pDebugControl->Output(DEBUG_OUTPUT_NORMAL, "Number of pages written: %u, length %u \n", pagesWritten, length);
			pDebugControl->Release();
		}
		CloseHandle(hFile);
	}

	return hRes;
}