#include "process.h"


PROCESS::PROCESS()
{
	eProcess = new ULONG64;
	activeProcessList = new ULONG;
	lockedPagesList = new ULONG64;
	lockedPagesCount = new ULONG;
	name = "";
	valid = false;
}

PROCESS::~PROCESS()
{
	delete[] eProcess;
	delete[] activeProcessList;
	delete[] lockedPagesList;
	delete[] lockedPagesCount;
}

PROCESS::PROCESS(IDebugClient* pDebugClient, PULONG64 addr)
{
	eProcess = new ULONG64;
	activeProcessList = new ULONG;
	lockedPagesList = new ULONG64;
	lockedPagesCount = new ULONG;
	valid = false;
	name = "";
	*eProcess = *addr;
	valid = readProcessList(pDebugClient);
	if(isProcessValid (pDebugClient) )
	{
		valid = readLockedPagesList(pDebugClient);
		valid = readLockedPagesCount(pDebugClient);
		valid = readProcessName(pDebugClient);
	}
	else
	{
		lockedPagesList = 0;
		lockedPagesCount = 0;
		valid = false;
	}
}

BOOL PROCESS::readProcessList(IDebugClient *pDebugClient)
{
	BOOL success = false;
	if( SUCCEEDED (FindFieldOffset(pDebugClient,"ActiveProcessLinks","_EPROCESS", "ntdll", activeProcessList)))
	{
		success = true;
	}
	else
	{
		*activeProcessList = 0;
	}
	return success;
}

BOOL PROCESS::readLockedPagesList(IDebugClient *pDebugClient)
{
	BOOL success = false;
	ULONG *offset = new ULONG;
	PVOID *buffer = new PVOID;
	if( SUCCEEDED (FindFieldOffset(pDebugClient,"LockedPagesList","_EPROCESS", "ntdll", offset)))
	{
		ULONG64 pageListOffset = *offset + *eProcess;
		if( SUCCEEDED (ReadVirtual(pDebugClient, pageListOffset, buffer, 8, NULL)))
		{
			success = true;
			*lockedPagesList = (ULONG64) *buffer;
		}
	}
	else
	{
		*lockedPagesList = 0;
	}
	delete[] offset;
	delete[] buffer;
	return success;
}

BOOL PROCESS::isProcessValid(IDebugClient *pDebugClient)
{
	//Check for valid unique ID:
	PVOID *uID = new PVOID;
	ULONG *offset = new ULONG;
	
	BOOL isValid = false;
	if( SUCCEEDED (FindFieldOffset(pDebugClient, "UniqueProcessId", "_EPROCESS", "ntdll", offset)))
	{
		ULONG64 uidOffset = *eProcess + *offset;
		if( SUCCEEDED (ReadVirtual(pDebugClient, uidOffset, uID, 8, NULL)))
		{
			if(*uID > 0)
			{
				isValid = true;
			}
		}
	}
	delete[] uID;
	delete[] offset;
	return isValid;

}

BOOL PROCESS::readLockedPagesCount(IDebugClient *pDebugClient)
{
	BOOL success = false;
	ULONG64 countOffset = 8;
	ULONG64 countAddr = *lockedPagesList + countOffset;
	PVOID *buf = new PVOID;

	HRESULT hr = ReadVirtual(pDebugClient, countAddr, buf, 8, NULL);
	if(hr == S_OK)
	{
		success = true;
		*lockedPagesCount = (ULONG) *buf;
	}

	delete[] buf;
	return success;
}

BOOL PROCESS::hasLockedPages()
{
	BOOL hasPages = false;
	if(*lockedPagesCount > 0)
	{
		hasPages = true;
	}
	return hasPages;
}

ULONG64 PROCESS::getNextProcess(IDebugClient *pDebugClient)
{
	ULONG64 nextOffset= *eProcess + (ULONG64) *activeProcessList;
	ULONG64 nextProcess = 0;
	PVOID* procBuffer = new PVOID;
	if( SUCCEEDED (ReadVirtual(pDebugClient, nextOffset, procBuffer, 8, NULL)))
	{
		nextProcess = (ULONG64) *procBuffer - *activeProcessList;
	}

	delete[] procBuffer;
	return nextProcess;
}

VOID PROCESS::update(IDebugClient *pDebugClient, PULONG64 addr)
{
	*eProcess = *addr;
	valid = readProcessList(pDebugClient);
	valid = readLockedPagesList(pDebugClient);
	valid = readLockedPagesCount(pDebugClient);
	valid = readProcessName(pDebugClient);	
}

BOOL PROCESS::readProcessName(IDebugClient *pDebugClient)
{

	PVOID *proName = new PVOID;
	ULONG *offset = new ULONG;
	ULONG nameLength = 15;
	BOOL success = false;

	if( SUCCEEDED (FindFieldOffset(pDebugClient, "ImageFileName", "_EPROCESS", "ntdll", offset)))
	{
		ULONG64 nameOffset = *eProcess + *offset;
		if( SUCCEEDED (ReadVirtual(pDebugClient, nameOffset, proName, nameLength, NULL)))
		{
			success = true;
			name = (PCSTR) proName;
		}
	}
	delete[] offset;
	return success;
}