#include "ConsoleFunc.h"
#include "MiscFunc.h"
#include "SupportFunc.h"
#include "MemoryFunc.h"
#include "SecureKernelFunc.h"

UINT numChars = 0;
const UINT MAXPRESS = 128;
TCHAR charBuffer[MAXPRESS]; //Max number of recorded keyboard presses
HANDLE hFil1, hFil2;
HANDLE hIn, hOut;
ULONG64 defaultPageDir = 0x000000000034A000;

UINT MainConsoleLoop(HANDLE hFile1, HANDLE hFile2)
{
	//Set console encoding:
	SetConsoleOutputCP(1252);

	//Init handles:
	hFil1 = hFile1;
	hFil2 = hFile2;
	hIn = GetStdHandle(STD_INPUT_HANDLE);
	hOut = GetStdHandle(STD_OUTPUT_HANDLE);

	BOOL quit = false;
	UINT returnValue = 0;
	_tprintf(_T("Welcome to the main program, please input the desired command. Type !quit to quit. \n"));
	_tprintf(_T(">"));

	//Main input loop:
	if( hIn != INVALID_HANDLE_VALUE && hOut != INVALID_HANDLE_VALUE)
	{
		while(!quit)
		{
			if(!ReadInput())
			{
				_tprintf(_T("Quitting! \n"));
				quit = true;
			}
			
		}
		CloseHandle(hIn);
		CloseHandle(hOut);
	}
	else
	{
		returnValue = -1;
	}

	return returnValue;
}

BOOL ConvertCommand(TCHAR *commandBuf, BYTE *command, DWORD charsRead)
{
	BOOL success = false;
	UINT iterations = (UINT) charsRead - 2;
	for(UINT i = 0; i < iterations; i++)
	{
		command[i] = (BYTE) commandBuf[i];
	}
	command[iterations] = 0;
	success = true;
	return success;
}

BOOL ReadInput()
{
	BOOL success = false;
	const DWORD length = 256;
	INPUT_RECORD irBuf[length];	
	DWORD eventsRead = 0;
	
	if(FlushConsoleInputBuffer(hIn))
	{
		ReadConsoleInput(hIn, irBuf, length, &eventsRead); 
		for(UINT i = 0; i < eventsRead; i++)
		{
			switch(irBuf[i].EventType)
			{
				case KEY_EVENT:
				if(!KeyEventProc(irBuf[i].Event.KeyEvent))
				{
					success = true;
				}
				break;

				default:
				success = true;
				break;
			}

		}
	}

	return success;
}

BOOL KeyEventProc(KEY_EVENT_RECORD keyRec)
{
	BOOL quit = false;
	if(keyRec.bKeyDown)
	{
		WORD keyCode = keyRec.wVirtualKeyCode;
		if(ValidKey(keyCode) || keyCode == VK_SPACE)
		{
			if(numChars < MAXPRESS)
			{
				TCHAR press = keyRec.uChar.UnicodeChar;
				_tprintf(_T("%c"),press);
				charBuffer[numChars] = press;
				numChars++;
			}
			else
			{
				numChars = 0;
			}
		}
		else if(keyCode == VK_RETURN)
		{
			_tprintf(_T("\n"));
			charBuffer[numChars] = 0;
			if(CheckCommand())
			{
				quit = true;
			}
		}
		else if(keyCode == VK_BACK)
		{
			if(numChars > 0)
			{
				numChars--;
				charBuffer[numChars] = 0;
				ClearLine();
			}
		}
	}
	return quit;
}

BOOL CheckCommand()
{
	BOOL quit = false;
	if(numChars > 0)
	{
		const UINT MAXARGUMENTS = 10;
		TCHAR *args[MAXARGUMENTS];
		UINT *positions = new UINT[numChars-1];
		UINT numArg = NumberOfArgs(positions);
		UINT *argLength = new UINT[numArg];
		if(numArg < MAXARGUMENTS)
		{
			InitArgs(positions, numArg, args, argLength);
			_tprintf(_T("Number of arguments: %u \n"), numArg);
			UINT commandLength = 0;
			if(numArg > 0)
			{
				commandLength = positions[0] - 1;
			}
			else
			{
				commandLength = numChars;
			}
			TCHAR *command = new TCHAR[commandLength];
			ReturnCommand(command, commandLength);
			_tprintf(_T("Command: %s.\n"),command);

			//Main command test loop:
			if( CSTR_EQUAL == CompareString(LOCALE_SYSTEM_DEFAULT, NORM_IGNORECASE, command, commandLength, _T("!quit"), 5))
			{
				_tprintf(_T("\n Goodbye! \n"));
				quit = true;
			}

			else if( CSTR_EQUAL == CompareString(LOCALE_SYSTEM_DEFAULT,NORM_IGNORECASE, command, commandLength, _T("!processlist"), 12))
			{
				ListProcesses(0x44FC7840, hFil1, hFil2, defaultPageDir);
			}

			else if( CSTR_EQUAL == CompareString(LOCALE_SYSTEM_DEFAULT,NORM_IGNORECASE, command, commandLength, _T("!translate"), 10))
			{
				if(numArg > 0)
				{
					ULONG64 virtAddress = StringToUlong64(args[0], argLength[0]);
					ULONG64 physAddress = 0;
					if( physAddress = TranslateAddress(virtAddress, defaultPageDir, hFil1, hFil2))
					{
						_tprintf(_T("The address %p , has physical address %p \n"), virtAddress, physAddress);
					}
					else
					{
						_tprintf(_T("Failed to translate address \n"));
					}
				}
			}

			else if( CSTR_EQUAL == CompareString(LOCALE_SYSTEM_DEFAULT,NORM_IGNORECASE, command, commandLength, _T("!setcontext"), 11))
			{
				if(numArg > 0)
				{
					ULONG64 newPageDir = StringToUlong64(args[0], argLength[0]);
					_tprintf(_T("Pagedir updated, old page dir: %p, new page dir: %p \n"), defaultPageDir, newPageDir);
					defaultPageDir = newPageDir;					
				}

			}

			else if( CSTR_EQUAL == CompareString(LOCALE_SYSTEM_DEFAULT, NORM_IGNORECASE, command, commandLength, _T("!peb"), 4))
			{
				if(numArg > 0)
				{
					ULONG64 pebAdr = StringToUlong64(args[0], argLength[0]);
					DumpPEB(pebAdr, defaultPageDir, hFil1, hFil2);
				}
			}

			else if( CSTR_EQUAL == CompareString(LOCALE_SYSTEM_DEFAULT, NORM_IGNORECASE, command, commandLength, _T("!process"), 8))
			{
				if(numArg > 0)
				{
					ULONG64 eProcessAdr = StringToUlong64(args[0], argLength[0]);
					ProcessData(eProcessAdr, hFil1, hFil2, defaultPageDir);
				}
			}

			else if ( CSTR_EQUAL == CompareString(LOCALE_SYSTEM_DEFAULT, NORM_IGNORECASE, command, commandLength, _T("!search"), 8))
			{
				TCHAR szFileName[MAX_PATH] = TEXT("d:\\temp\\TestWrite.dat");
				ULONG64 offsets[4096];
				_tprintf(_T("Searching for Process Headers:\n"));
				UINT totalMatches = SearchForProcHeader(hFil1, 1999, offsets);
				_tprintf(_T("\nProcess Image Names: \n"));
				SearchForImageNames(offsets, totalMatches, szFileName);
			}

			else if ( CSTR_EQUAL == CompareString(LOCALE_SYSTEM_DEFAULT, NORM_IGNORECASE, command, commandLength, _T("!skvars"), 7))
			{
				skDumpVariables(hFil2, 0x024ae000);

			}

			else if ( CSTR_EQUAL == CompareString(LOCALE_SYSTEM_DEFAULT, NORM_IGNORECASE, command, commandLength, _T("!skvar"), 6))
			{
				if(numArg > 0)
				{
					ULONG64 offset = StringToUlong64(args[0], argLength[0]);
					skDumpVar(hFil2, 0x024ae000, offset);
				}
			}

			else if ( CSTR_EQUAL == CompareString(LOCALE_SYSTEM_DEFAULT, NORM_IGNORECASE, command, commandLength, _T("!dv"), 3))
			{
				if(numArg == 1)
				{
					ULONG64 virtAdr = StringToUlong64(args[0], argLength[0]);
					OutputVirtual(virtAdr, defaultPageDir, hFil1, hFil2, 8);
				}
				else if(numArg > 1)
				{
					ULONG64 virtAdr = StringToUlong64(args[0], argLength[0]);
					ULONG64 length = StringToUlong64(args[1], argLength[1]);
					OutputVirtual(virtAdr, defaultPageDir, hFil1, hFil2, (UINT) length);
				}
			}

			else if ( CSTR_EQUAL == CompareString(LOCALE_SYSTEM_DEFAULT, NORM_IGNORECASE, command, commandLength, _T("!dp"), 3))
			{
				if(numArg == 1)
				{
					ULONG64 physAdr = StringToUlong64(args[0], argLength[0]);
					OutputPhysical(physAdr, hFil1, hFil2, 8);
				}
				else if(numArg > 1)
				{
					ULONG64 physAdr = StringToUlong64(args[0], argLength[0]);
					ULONG64 length = StringToUlong64(args[1], argLength[1]);
					OutputPhysical(physAdr, hFil1, hFil2, (UINT) length);
				}
			}

			else if ( CSTR_EQUAL == CompareString(LOCALE_SYSTEM_DEFAULT, NORM_IGNORECASE, command, commandLength, _T("!add"), 4))
			{
				if(numArg > 1)
				{
					ULONG64 ar1 = StringToUlong64(args[0], argLength[0]);
					ULONG64 ar2 = StringToUlong64(args[1], argLength[1]);
					_tprintf(_T("Result: 0x%p"),ar1+ar2); 

				}
			}
			else if ( CSTR_EQUAL == CompareString(LOCALE_SYSTEM_DEFAULT, NORM_IGNORECASE, command, commandLength, _T("!sub"), 4))
			{
				if(numArg > 1)
				{
					ULONG64 ar1 = StringToUlong64(args[0], argLength[0]);
					ULONG64 ar2 = StringToUlong64(args[1], argLength[1]);
					_tprintf(_T("Result: 0x%p"),ar1-ar2); 

				}
			}
			else if ( CSTR_EQUAL == CompareString(LOCALE_SYSTEM_DEFAULT, NORM_IGNORECASE, command, commandLength, _T("!skob"), 5))
			{
				if(numArg > 0)
				{
					ULONG64 ar1 = StringToUlong64(args[0], argLength[0]);
					SearchForSKOB(hFil1, hFil2, ar1);
				}
			}
			else if ( CSTR_EQUAL == CompareString(LOCALE_SYSTEM_DEFAULT, NORM_IGNORECASE, command, commandLength, _T("!sklm"), 5))
			{
				if(numArg > 0)
				{
					ULONG64 ar1 = StringToUlong64(args[0], argLength[0]);
					skLoadedModules(ar1, defaultPageDir, hFil1, hFil2);
				}
			}
			else if ( CSTR_EQUAL == CompareString(LOCALE_SYSTEM_DEFAULT, NORM_IGNORECASE, command, commandLength, _T("!vad"), 4))
			{
				if(numArg > 0)
				{
					ULONG64 ar1 = StringToUlong64(args[0], argLength[0]);
					TraverseVadTree(ar1, hFil1, hFil2, defaultPageDir);
				}
			}
			else
			{
				_tprintf(_T("\n Not a valid command \n"));
			}
		}
		else
		{
			_tprintf(_T("ERROR: Too many arguments or spaces \n"));
		}
	}

	_tprintf(_T("\n>"));
	numChars = 0;
	return quit;
}

BOOL ValidKey(WORD keyCode)
{
	BOOL valid = false;
	if( keyCode >= 0x30 && keyCode <= 0x5a)
	{
		valid = true;
	}
	return valid;
}

BOOL ClearLine()
{
	BOOL success = false;
	
	CONSOLE_SCREEN_BUFFER_INFO csbi; 

	if(GetConsoleScreenBufferInfo(hOut, &csbi))
	{
		COORD currentPosition = csbi.dwCursorPosition;
		if(currentPosition.X > 0)
		{
			COORD newPosition = {currentPosition.X - 1, currentPosition.Y};
			DWORD charsWritten = 0;
			if(FillConsoleOutputCharacter(hOut, (TCHAR) ' ', numChars+2, newPosition, &charsWritten))
			{
				if(SetConsoleCursorPosition(hOut, newPosition))
				{
					success = true;
				}
			}
		}
	}
	
	return success;
}

UINT NumberOfArgs(UINT *positions)
{
	UINT numArgs = 0;
	UINT *posBuf = new UINT[numChars];

	for(UINT i = 1; i < numChars -1; i++)
	{
		if(charBuffer[i] == (TCHAR) ' ' )
		{
			posBuf[numArgs] = i+1;
			numArgs++;
		}

	}
	for(UINT i = 0; i < numArgs; i++)
	{
		positions[i] = posBuf[i];

	}
	delete[] posBuf;
	return numArgs;
}


BOOL InitArgs(UINT *positions, UINT numArgs, TCHAR *args[], UINT *argLength)
{
	BOOL success = false;
	for(UINT i = 0; i < numArgs; i++)
	{
		UINT cp = positions[i]; //CurrentPosition
		UINT stringLen = 0;
		if(i < numArgs - 1)
		{
			stringLen = positions[i+1] - cp - 1;
		}
		else
		{
			stringLen = numChars - cp;
		}
		TCHAR *arg = new TCHAR[stringLen+1];
		for(UINT j = 0; j < stringLen; j++)
		{
			arg[j] = charBuffer[cp+j];			
		}
		arg[stringLen] = 0;
		argLength[i] = stringLen;
		args[i] = arg;
		success = true;
	}
	return success;
}


BOOL ReturnCommand(TCHAR *command, UINT commandLength)
{
	for(UINT i = 0; i < commandLength; i++)
	{
		command[i] = charBuffer[i];
	}
	command[commandLength] = 0;
	return true;
}