#include "MemoryFunc.h"
#include "PageEntry.h"
#include "SupportFunc.h"

ULONG64 TranslateAddress(ULONG64 virtAddress, ULONG64 pageDir, HANDLE hFile1, HANDLE hFile2)
{	
	HANDLE hOpen;
	ULONG64 physicalAddress = 0;
	ULONG64 pgDir = pageDir;
	
	//PXE:
	PVOID *pxeCont = new PVOID;
	ULONG64 pxeAdr = 0;
	hOpen = ChooseFile(&hFile1, &hFile2, &pgDir);
	pxeAdr = TranslatePXE(virtAddress, pgDir);
	ReadPhysical(pxeAdr, pxeCont, 8, hOpen);
	//_tprintf(_T("PXE cont: %p \n"), *pxeCont);

	//PPE:
	PageEntry *ppe = new PageEntry( (ULONG64) *pxeCont);
	PVOID *ppeCont = new PVOID;
	ULONG64 ppeAdr = 0;
	ULONG64 pAdr = ppe->physAddress();
	hOpen = ChooseFile(&hFile1, &hFile2, &pAdr);
	ppeAdr = TranslatePPE(virtAddress, pAdr);
	ReadPhysical(ppeAdr, ppeCont, 8, hOpen);
		
	//_tprintf(_T("PPE cont: %p \n"), *ppeCont);

	//PDE:
	PageEntry *pde = new PageEntry( (ULONG64) *ppeCont);
	PVOID *pdeCont = new PVOID;
	ULONG64 pdeAdr = 0;
	pAdr = pde->physAddress();
	hOpen = ChooseFile(&hFile1, &hFile2, &pAdr);
	pdeAdr = TranslatePDE(virtAddress, pAdr);
	ReadPhysical(pdeAdr, pdeCont, 8, hOpen);
	
	//_tprintf(_T("PDE cont: %p \n"), *pdeCont);

	//Physical Address:
	PageEntry *pte = new PageEntry ( (ULONG64) *pdeCont);
	if(pte->isLarge())
	{
		physicalAddress = TranslateLargePage(virtAddress, pte->physAddress());
	}
	else
	{
		PVOID *pteCont = new PVOID;
		ULONG64 pteAdr = 0;
		pAdr = pte->physAddress();
		hOpen = ChooseFile(&hFile1, &hFile2, &pAdr);
		pteAdr = TranslatePTE(virtAddress, pAdr);
		ReadPhysical(pteAdr, pteCont, 8, hOpen);
		PageEntry *phys = new PageEntry ( (ULONG64) *pteCont);
		//_tprintf(_T("PTE cont: %p \n"), *pteCont); 

		physicalAddress = phys->physAddress() + (virtAddress & 0xFFF);
	}
	//_tprintf(_T("Physical Address: %p \n"), physicalAddress);
	

	//Cleanup:
	delete[] pxeCont;
	delete[] ppeCont;
	delete[] pdeCont;
	delete ppe;
	delete pde;
	delete pte;

	return physicalAddress;

}

BOOL ReadPhysical(ULONG64 physAddress, PVOID buf, UINT bytes, HANDLE hFil)
{
	BOOL success = false;
	LARGE_INTEGER distance;
	distance.QuadPart = physAddress;
	if(SetFilePointerEx(hFil, distance, NULL, FILE_BEGIN))
	{
		DWORD bytesRead = 0;
		if(ReadFile(hFil, buf, bytes, &bytesRead, NULL))
		{
			if(bytes == (UINT) bytesRead)
			{
				success = true;
			}
			else
			{
				_tprintf(_T("ERROR: Buffer Mismatch \n"));
			}
		}
		else
		{
			_tprintf(_T("Invalid address: %p \n"),physAddress);
		}
		if(!ReturnFileHandleToStart(hFil))
		{
			success = false;
		}
	}
	return success;
}

BOOL ReadVirtual(ULONG64 virtAddress, ULONG64 pageDir, HANDLE hFile1, HANDLE hFile2, PVOID buf, UINT bytes)
{
	ULONG64 physAdr = TranslateAddress(virtAddress, pageDir, hFile1, hFile2);
	HANDLE hOpen = ChooseFile(&hFile1, &hFile2, &physAdr);
	BOOL success = ReadPhysical(physAdr, buf, bytes, hOpen);
	return success;
}

ULONG64 TranslatePXE(ULONG64 virtAddress, ULONG64 pagedir)
{
	const ULONG64 PXEmask = 0xFF8000000000;
	ULONG64 offset = (virtAddress & PXEmask) >> 39;
	return pagedir+(offset*8);
}

ULONG64 TranslatePPE(ULONG64 virtAddress, ULONG64 physAdr)
{
	const ULONG64 PPEmask = 0x7FC0000000;
	ULONG64 offset = (virtAddress & PPEmask) >> 30;
	return physAdr+(offset*8);
}

ULONG64 TranslatePDE(ULONG64 virtAddress, ULONG64 physAdr)
{
	const ULONG64 PDEmask = 0x3FE00000;
	ULONG64 offset = (virtAddress & PDEmask) >> 21;
	return physAdr+(offset*8);
}

ULONG64 TranslatePTE(ULONG64 virtAddress, ULONG64 physAdr)
{
	const ULONG64 PTEmask = 0x1FF000;
	ULONG64 offset = (virtAddress & PTEmask) >> 12;
	return physAdr+(offset*8);
}



ULONG64 TranslateLargePage(ULONG64 virtAddress, ULONG64 physAdr)
{
	const ULONG64 PTEmask = 0x1FF000;
	const ULONG64 pageMask = 0xFFF;
	ULONG64 offset = (virtAddress & PTEmask);
	ULONG64 pageOffset = (virtAddress & pageMask);
	return (physAdr + offset + pageOffset);
}


HANDLE ChooseFile(HANDLE *hFile1, HANDLE *hFile2, ULONG64 *physAdr)
{
	const ULONG64 THRESHOLD = 0x100000000;
	const ULONG64 MASK      = 0xFFFFFFFF;
	HANDLE hFil = *hFile1;
	if(*physAdr < THRESHOLD)
	{
		hFil = *hFile2;
	}
	*physAdr = *physAdr & MASK;
	return hFil;
}