#include "MiscFunc.h"
#include "MemoryFunc.h"


ULONG64 ConvertToUlong64(BYTE *b)
{
	LARGE_INTEGER large;
	large.LowPart = (b[3] << 24) | (b[2] << 16) | (b[1] << 8) | (b[0]);
	large.HighPart = (b[7] << 24) | (b[6] << 16) | (b[5] << 8) | (b[4]);
	ULONG64 number = large.QuadPart;
	return number;
}

UINT ConvertToUint(BYTE *b)
{
	UINT num = (b[3] << 24) | (b[2] << 16) | (b[1] << 8) | (b[0]);
	return num;
}

VOID ReverseUlong64(BYTE *b)
{
	UINT end = sizeof(ULONG64) / 2;
	UINT endOffset = end - 1;
	for(UINT i = 0; i < end; i++)
	{
		BYTE swap = b[endOffset];
		b[endOffset] = b[i];
		b[i] = swap;
		endOffset--;
	}
	return;
}

VOID DwordToByteArr(BYTE *arr, DWORD num, UINT off)
{
	//Bitmasks:
	ULONG64 bitMask = 0xFF;

	UINT startShift = 24;
	for(UINT i = off; i < sizeof(DWORD); i++)
	{
		arr[i] = ((( bitMask << startShift) & num ) >> startShift );
		startShift -= 8;
	}

	return;
}

VOID Ulong64ToByteArr(BYTE *arr, ULONG64 num, UINT off)
{
	ULONG64 bitMask = 0xFF;

	ULONG64 ulong64Length = sizeof(ULONG64);
	UINT startShift = 56;
	for(UINT i = off; i < ulong64Length; i++)
	{
		arr[i] = ((( bitMask << startShift) & num ) >> startShift );
		startShift -= 8;
	}
	return;

}

ULONG64 ReadPointer(BYTE *b, UINT offset)
{
	ULONG64 param = 0;
	const UINT POINTSIZE = 8;
	UINT counter = POINTSIZE;
	BYTE buf[POINTSIZE];
		
	for(UINT i = 0; i < POINTSIZE; i++)
	{
		buf[i] = b[i + offset];
	}
	param = ConvertToUlong64(buf);
	return param;
}

UINT ReadUint(BYTE *b, UINT offset)
{
	UINT param = 0;
	const UINT POINTSIZE = 4;
	UINT counter = POINTSIZE;
	BYTE buf[POINTSIZE];
		
	for(UINT i = 0; i < POINTSIZE; i++)
	{
		buf[i] = b[i + offset];
	}
	param = ConvertToUint(buf);
	return param;
}

BOOL PrintUnicodeString(ULONG64 physAdr, ULONG64 data, HANDLE hOpen)
{
	//Read Length:
	UINT length = data & 0xFFFF;
	UINT maxLength = ((data & 0xFFFF0000) >> 16);
	BOOL success = false;

	//Allocate memory:
	__try{
		TCHAR *string = new TCHAR[maxLength];
	
		if(success = ReadPhysical(physAdr, string, maxLength, hOpen))
		{
			_tprintf(_T("%s \n"), string);
		}
		else
		{
			_tprintf(_T("Failed to read string \n"));
		}
		//delete[maxLength] string;
	}
	__except (EXCEPTION_EXECUTE_HANDLER){

		_tprintf(_T("Memory allocation error \n"));
	}
	return success;

}

BOOL ShortenString(TCHAR *inString, TCHAR *outString, UINT length)
{
	BOOL success = true;
	for(UINT i = 0; i < length; i++)
	{
		outString[i] = inString[i];
	}
	return success;
}

ULONG64 IntegerPower(UINT exponent, ULONG base)
{
	ULONG64 number = 1;
	ULONG64 bs = (ULONG64) base;
	for(UINT i = 0; i < exponent; i++)
	{
		number *= bs;
	}
	//_tprintf(_T("Multiplier %p \n"), number);
	return number;
}

ULONG64 StringToUlong64(TCHAR *string, UINT len)
{
	//Test if the string is a hexidecimal number (prefixed by 0x).
	BOOL hexaDecimal = false;
	UINT startingPoint = 0;
	if(len > 2)
	{
		if( string[0] == (TCHAR) '0' && string[1] == (TCHAR) 'x')
		{
			hexaDecimal = true;
			startingPoint = 1;
		}
	}
	
	UINT base = 10;
	if(hexaDecimal)
	{
		base = 16;
	}

	ULONG64 stringNumber = 0;
	UINT exponent = 0;
	for(UINT i = len - 1; i > startingPoint; i--)
	{
		TCHAR activeChar = string[i];
		ULONG num = NormalizeChar(activeChar);
		stringNumber += IntegerPower(exponent, base) * (ULONG64)  num;
		exponent++;
	}
	_tprintf(_T("\nInput string %s, output number %p \n"), string, stringNumber);
	return stringNumber;
}

ULONG NormalizeChar(TCHAR numberChar)
{
	//Numerical char:
	ULONG value = 0;
	ULONG number = (ULONG) numberChar;
	if(number > 47 && number < 58)
	{
		value = number - 48;
	}
	//Hex (A-F) char:
	else if(number > 64 && number < 71)
	{
		value = number - 55;
	}
	//Hex (a-f) char:
	else if(number > 96 && number < 103)
	{
		value = number - 87;
	}
	else
	{
		value = 0;
	}
	//_tprintf(_T("Normalized value: %u "), value);
	return value;
}
BOOL ValidAsciiChar(CHAR asciiChar)
{
	BOOL valid = false;
	if(asciiChar > 0x20 && asciiChar < 0x7f)
	{
		valid = true;
	}
	return valid;
}
