#include "PEB.h"
#include "MemoryFunc.h"
#include "MiscFunc.h"

PEB::PEB()
{
	start = 0;
	processParam = 0;
	valid = false;
}

PEB::PEB(ULONG64 startAdr)
{
	start = startAdr;
	processParam = 0;
	valid = false;
}

PEB::~PEB()
{
}

BOOL PEB::isValid()
{
	return valid;
}

BOOL PEB::ReadData(HANDLE hFile1, HANDLE hFile2, ULONG64 pageDir)
{
	BOOL success = false;
	ULONG64 physAdr = TranslateAddress(start, pageDir, hFile1, hFile2);
	HANDLE hOpen = ChooseFile(&hFile1, &hFile2, &physAdr);

	if(ReadPhysical(physAdr, rawData, PEBSIZE, hOpen))
	{
		success = true;
		valid = true;
		processParam = ReadProcessParam();
		valid = ReadParamData(hFile1, hFile2, pageDir);
	}
	return success;
}

BOOL PEB::ReadParamData(HANDLE hFile1, HANDLE hFile2, ULONG64 pageDir)
{
	BOOL success = false;
	if(valid)
	{
		ULONG64 physAdr = TranslateAddress(processParam, pageDir, hFile1, hFile2);
		HANDLE hOpen = ChooseFile(&hFile1, &hFile2, &physAdr);

		if(ReadPhysical(physAdr, paramData, PARAMSIZE, hOpen))
		{
			success = true;
		}
	}
	return success;
}

ULONG64 PEB::ReadProcessParam()
{
	ULONG64 param = 0;
	if(valid)
	{
		
		param = ReadPointer(rawData, PAROFF);
		
	}
	return param;
}



ULONG64 PEB::getProcParam()
{
	return processParam;
}

/*3: kd> dt _RTL_USER_PROCESS_PARAMETERS
ntdll!_RTL_USER_PROCESS_PARAMETERS
   +0x000 MaximumLength    : Uint4B
   +0x004 Length           : Uint4B
   +0x008 Flags            : Uint4B
   +0x00c DebugFlags       : Uint4B
   +0x010 ConsoleHandle    : Ptr64 Void
   +0x018 ConsoleFlags     : Uint4B
   +0x020 StandardInput    : Ptr64 Void
   +0x028 StandardOutput   : Ptr64 Void
   +0x030 StandardError    : Ptr64 Void
   +0x038 CurrentDirectory : _CURDIR
   +0x050 DllPath          : _UNICODE_STRING
   +0x060 ImagePathName    : _UNICODE_STRING
   +0x070 CommandLine      : _UNICODE_STRING
   +0x080 Environment      : Ptr64 Void
   +0x088 StartingX        : Uint4B
   +0x08c StartingY        : Uint4B
   +0x090 CountX           : Uint4B
   +0x094 CountY           : Uint4B
   +0x098 CountCharsX      : Uint4B
   +0x09c CountCharsY      : Uint4B
   +0x0a0 FillAttribute    : Uint4B
   +0x0a4 WindowFlags      : Uint4B
   +0x0a8 ShowWindowFlags  : Uint4B
   +0x0b0 WindowTitle      : _UNICODE_STRING
   +0x0c0 DesktopInfo      : _UNICODE_STRING
   +0x0d0 ShellInfo        : _UNICODE_STRING
   +0x0e0 RuntimeData      : _UNICODE_STRING
   +0x0f0 CurrentDirectores : [32] _RTL_DRIVE_LETTER_CURDIR
   +0x3f0 EnvironmentSize  : Uint8B
   +0x3f8 EnvironmentVersion : Uint8B
   +0x400 PackageDependencyData : Ptr64 Void
   +0x408 ProcessGroupId   : Uint4B
   +0x40c LoaderThreads    : Uint4B*/

VOID PEB::OutPutData(HANDLE hFile1, HANDLE hFile2, ULONG64 pageDir)
{
	if(valid)
	{
		//Output PEB data pointers:
		_tprintf(_T("\nPEB data: \n"));
		_tprintf(_T("PEB virtual address: %p \n"),start);
		_tprintf(_T("Process Parameter virtual address: %p \n"),processParam);
		ULONG64 environmentPoint = ReadPointer(paramData, 0x80);
		_tprintf(_T("Environment block: %p \n"),environmentPoint);
		_tprintf(_T("Environment block physical address: %p \n"),TranslateAddress(environmentPoint, pageDir, hFile1, hFile2));

		_tprintf(_T("\nProcess Parameter Data: \n"));

		//Output ImagePathName:
		ULONG64 szData = ReadPointer(paramData, 0x60);
		ULONG64 szPointer  = ReadPointer(paramData, 0x68);
		ULONG64 physAdr = 0;
		HANDLE hOpen = INVALID_HANDLE_VALUE;
		if(szPointer != NULL)
		{
			physAdr = TranslateAddress(szPointer, pageDir, hFile1, hFile2);
			hOpen = ChooseFile(&hFile1, &hFile2, &physAdr);
			_tprintf(_T("ImagePathName: "));
			PrintUnicodeString(physAdr, szData, hOpen);
		}

		//Output Commandline:
		szData = ReadPointer(paramData, 0x70);
		szPointer  = ReadPointer(paramData, 0x78);
		if(szPointer != NULL)
		{
			physAdr = TranslateAddress(szPointer, pageDir, hFile1, hFile2);
			hOpen = ChooseFile(&hFile1, &hFile2, &physAdr);
			_tprintf(_T("CommandLine: "));
			PrintUnicodeString(physAdr, szData, hOpen);
		}

		//Output WindowsTitle:
		szData = ReadPointer(paramData, 0xb0);
		szPointer  = ReadPointer(paramData, 0xb8);
		if(szPointer != NULL)
		{
			physAdr = TranslateAddress(szPointer, pageDir, hFile1, hFile2);
			hOpen = ChooseFile(&hFile1, &hFile2, &physAdr);
			_tprintf(_T("WindowTitle: "));
			PrintUnicodeString(physAdr, szData, hOpen);
		}

		//Output DesktopInfo:
		szData = ReadPointer(paramData, 0xc0);
		szPointer  = ReadPointer(paramData, 0xc8);
		if(szPointer != NULL)
		{
			physAdr = TranslateAddress(szPointer, pageDir, hFile1, hFile2);
			hOpen = ChooseFile(&hFile1, &hFile2, &physAdr);
			_tprintf(_T("DesktopInfo: "));
			PrintUnicodeString(physAdr, szData, hOpen);
		}

		
	}
}