#include "SupportFunc.h"
#include "PoolHeader.h"
#include "MemoryFunc.h"
#include "PEB.h"
#include "MiscFunc.h"
#include "BALNODE.h"
#include <stack>
#include "VAD.h"

HANDLE FileOpen(TCHAR *szFileName)
{	
	HANDLE hFile = CreateFile(
					szFileName,
					GENERIC_READ,
					0,
					NULL,
					OPEN_EXISTING,
					FILE_ATTRIBUTE_NORMAL | FILE_FLAG_SEQUENTIAL_SCAN,
					NULL);

	return hFile;
}

BOOL ReturnFileHandleToStart(HANDLE hFile)
{
	BOOL success = false;
	LARGE_INTEGER start;
	start.QuadPart = 0;
	if(SetFilePointerEx(hFile, start, NULL, FILE_BEGIN))
	{
		success = true;
	}
	return success;
}


UINT SearchForProcHeader(HANDLE hOpen, UINT noMb, ULONG64 *offsets)
{
	HANDLE hFile = hOpen;
	//BOOL success = false;
	const DWORD DWPAGESIZE = 4096;
	const UINT patLength = 7;
	UINT iterations = (1024*1024)*noMb / DWPAGESIZE;
	UINT *ReadBytes = new UINT;
	ULONG64 totalOffset = 0;
	UINT totalMatches = 0;
	ULONG64 matchOffsets[DWPAGESIZE];
	*ReadBytes = 0;
	BYTE buffer[DWPAGESIZE];
	BYTE pattern[patLength];
	//Set pattern:
	pattern[0] = 0x00; //Index
	pattern[1] = 0x84; //// 0x84; //Blocksize
	pattern[2] = 0x02; //Pooltype
	pattern[3] = 0x50; //Tag (Proc)50 72 6f 63
	pattern[4] = 0x72;
	pattern[5] = 0x6f;
	pattern[6] = 0x63;

	if(hFile != INVALID_HANDLE_VALUE)
	{
		for(UINT i = 0; i < iterations; i++)
		{
			//Read I/0 synchronously:
			if(SafeRead(hFile, buffer, DWPAGESIZE, ReadBytes))
			{
				//ULONG64 offsets[DWPAGESIZE];
				UINT matches = 0;
				
				if( matches = Match(buffer, pattern, matchOffsets, *ReadBytes, patLength))
				{
					for(UINT i = 0; i < matches; i++)
					{
						offsets[totalMatches] = matchOffsets[i] + totalOffset;
						totalMatches++;
						_tprintf( _T("Process header found at offset: %p \n"), matchOffsets[i] + totalOffset);
					}
				}
				else
				{
					//_tprintf( _T("Pattern not found. \n"));
				}
				totalOffset += *ReadBytes;
			}
		}
		CloseHandle(hFile);
	}
	delete[] ReadBytes;
	return totalMatches;
}

BOOL SafeRead(HANDLE hFile, BYTE *buffer, UINT desiredBytes, PUINT readBytes)
{
	LARGE_INTEGER *lpFileSize = new LARGE_INTEGER;
	BOOL success = false;
	DWORD dwBytesToRead = desiredBytes;
	DWORD dwBytesRead = 0;

	if( success = GetFileSizeEx(hFile, lpFileSize))
	{
		ULONG64 fileSize = lpFileSize->QuadPart;
		if(fileSize < desiredBytes)
		{
			dwBytesToRead = lpFileSize->LowPart;
		}
		success = ReadFile(hFile, buffer, dwBytesToRead, &dwBytesRead, NULL);
		*readBytes = (UINT) dwBytesRead;

		if(dwBytesToRead != dwBytesRead)
		{
			_tprintf(_T("Buffer mismatch error \n"));
			success = false;
		}
	}
	if(!success)
	{
		*readBytes = 0;
	}
	delete[] lpFileSize;
	return success;
}

UINT Match(BYTE *buffer, BYTE *pattern, PULONG64 offset, DWORD bufSize, UINT patLength)
{
	UINT matches = 0;
	UINT searchSpace = (UINT) bufSize;
	for(UINT i = 0; i < searchSpace; )
	{
		UINT j = 0; //offset into searchPattern
		BOOL valid = true;
		BOOL noMatch = true;
		while(valid && noMatch)
		{
			BYTE currChar = buffer[i+j]; 
			if( currChar == pattern[j])
			{
				j++;
				if(j == patLength)
				{

					noMatch = false;
					offset[matches] = i - 1;
					matches++;
					i += j;
				}
			}
			else
			{
				i += j + 1;
				j = 0;
				valid = false;
			}
		}
	}
	return matches;
}

BOOL ValidHeader(BYTE *buffer, UINT offset)
{
	BOOL valid = false;
	ULONG64 addr = offset-4;
	if(offset < 4)
	{
		//POOL *pHeader = new POOL(buffer, addr);
		//valid = pHeader->getValid();
	}
	return valid;
}

BOOL SearchForImageNames(ULONG64 *offsets, UINT matches, TCHAR *szFileName)
{
	BOOL success = false;
	HANDLE hFile = CreateFile(szFileName,
							GENERIC_READ,
							0,
							NULL,
							OPEN_EXISTING,
							FILE_ATTRIBUTE_NORMAL | FILE_FLAG_RANDOM_ACCESS,
							NULL);
	if(hFile != INVALID_HANDLE_VALUE)
	{
		for(UINT i = 0; i < matches; i++)
		{
			//Position file pointer:
			ULONG64 offsetImageName = 0x450 + 0x80; 
			LARGE_INTEGER distance;
			distance.QuadPart = offsets[i] + offsetImageName; 
			if(SetFilePointerEx(hFile, distance, NULL, FILE_BEGIN))
			{
				//Read ImageFileNameData:
				CHAR imageName[16];
				imageName[15] = 0;
				DWORD bytesRead = 0;
				ReadFile(hFile, imageName, 15, &bytesRead, NULL);
				//Output name:
				printf("Process Name: %s, EPROCESS offset: %p \n", imageName, offsets[i] + 0x80);

			}
			else
			{
				_tprintf(_T("Error setting file pointer offset %p \n"), offsets[i]);
			}
			
		}
		CloseHandle(hFile);
	}
	return success;				

}

BOOL ListProcesses(ULONG64 eprocessAdr, HANDLE hFile1, HANDLE hFile2, ULONG64 pageDir)
{
	BOOL success = false;
	HANDLE hOpen;
	ULONG64 currEprocess = 0;
	ULONG64 pageDirOffset = 0x28;
	ULONG64 activeProcOffset = 0x2f0;
	ULONG64 pebOffset = 0x3f8;
	ULONG64 imageNameOffset = 0x450;
	ULONG64 startEprocess = 0;

	//Read page dir:
	
	//PVOID *pageDir = new PVOID;
	//ReadPhysical(eprocessAdr + pageDirOffset, pageDir, 8, hFile1);

	//Read ActiveProcessLink:
	
	PVOID *processLink = new PVOID;
	ReadPhysical(eprocessAdr + activeProcOffset, processLink, 8, hFile1);

	ULONG64 currLink = (ULONG64) * processLink;
	startEprocess = currLink - activeProcOffset;
	BOOL once = true;
	while( startEprocess != currEprocess)
	{
		if(once)
		{
			currEprocess = startEprocess;
			once = false;
		}
		//Get next link:
		ULONG64 address = TranslateAddress(currLink, pageDir, hFile1, hFile2);
		hOpen = ChooseFile(&hFile1, &hFile2, &address);
		ReadPhysical(address, processLink, 8, hOpen);

		//Read name:
		CHAR imgName[15];		
		currLink = (ULONG64) *processLink;
		ReadPhysical(address - activeProcOffset + imageNameOffset, imgName, 15, hOpen);

		printf(("Curr EPROCESS: 0x%p, name %s. \n"), currEprocess, imgName);
		
		//Follow the next link:
		currEprocess = currLink - activeProcOffset;
		
	}
	_tprintf(_T("\n"));
	delete [] processLink;
	return success;
}

BOOL ProcessData(ULONG64 vEprocessAdr, HANDLE hFile1, HANDLE hFile2, ULONG64 pageDir)
{
	BOOL success = false;
	ULONG64 pageDirOffset = 0x28;
	ULONG64 securePidOffset = 0x2d0;
	ULONG64 pebOffset = 0x3f8;
	const UINT POINTERSIZE = 8;
	ULONG64 imageNameOffset = 0x450;
	HANDLE hOpen;

	ULONG64 eprocessBlock = TranslateAddress(vEprocessAdr, pageDir, hFile1, hFile2);
	hOpen = ChooseFile(&hFile1, &hFile2, &eprocessBlock);
	
	//Read page DIR:
	PVOID *processPageDir = new PVOID;
	ReadPhysical(eprocessBlock + pageDirOffset, processPageDir, POINTERSIZE, hOpen);

	//Read PEB:
	PVOID *peb = new PVOID;
	ReadPhysical(eprocessBlock + pebOffset, peb, POINTERSIZE, hOpen);

	//Read name:
	CHAR imgname[15];
	ReadPhysical(eprocessBlock + imageNameOffset, imgname, 15, hOpen);

	//Read Secure PID:
	PVOID *secPid = new PVOID;
	ReadPhysical(eprocessBlock + securePidOffset, secPid, POINTERSIZE, hOpen);

	//Output:
	printf("\nProcess: %p, (%s) \n", vEprocessAdr, imgname);
	_tprintf(_T("Page Dir: 0x%p \n"), (ULONG64) *processPageDir);
	_tprintf(_T("PEB: 0x%p \n"), (ULONG64) *peb);
	_tprintf(_T("Secure PID: 0x%p \n"), (ULONG64) *secPid);


	_tprintf(_T("\n"));

	//Cleanup:
	delete[] processPageDir;
	delete[] peb;
	delete[] secPid;
	
	return true;
}

BOOL OpenFiles(HANDLE *hFile1, HANDLE *hFile2, TCHAR *fileName1, TCHAR *fileName2)
{
	BOOL success = false;
	

	
	*hFile1 = CreateFile( fileName1,
							GENERIC_READ,
							0,
							NULL,
							OPEN_EXISTING,
							FILE_ATTRIBUTE_NORMAL | FILE_FLAG_RANDOM_ACCESS,
							NULL);
	*hFile2 = CreateFile( fileName2,
							GENERIC_READ,
							0,
							NULL,
							OPEN_EXISTING,
							FILE_ATTRIBUTE_NORMAL | FILE_FLAG_RANDOM_ACCESS,
							NULL);

	if(*hFile1 != INVALID_HANDLE_VALUE && *hFile2 != INVALID_HANDLE_VALUE)
	{
		success = true;

	}
	return success;
}

BOOL CloseFiles(HANDLE *hFile1, HANDLE *hFile2)
{
	BOOL success = false;
	if(*hFile1 != INVALID_HANDLE_VALUE && *hFile2 != INVALID_HANDLE_VALUE)
	{
		CloseHandle(*hFile1);
		CloseHandle(*hFile2);
		success = true;
	}
	return success;
}

BOOL DumpPEB(ULONG64 pebAdr, ULONG64 pageDir, HANDLE hFile1, HANDLE hFile2)
{
	BOOL success = false;
	PEB *pb = new PEB(pebAdr);
	pb->ReadData(hFile1, hFile2, pageDir);
	if(pb->isValid())
	{
		success = true;
		//_tprintf(_T("PEB read sucessfully, PROCESS PARAM VIRT adr: %p!\n"), pb->getProcParam());
		//_tprintf(_T("PROCESS PARAM Physical Address: %p \n"), TranslateAddress(pb->getProcParam(), pageDir, hFile1, hFile2));
		pb->OutPutData(hFile1, hFile2, pageDir);
	}
	else
	{
		_tprintf(_T("Failed to read PEB \n"));
	}
	delete pb;
	return success;
	
}

BOOL OutputVirtual(ULONG64 virtAddress, ULONG64 pageDir, HANDLE hFile1, HANDLE hFile2, UINT len)
{
	ULONG64 physAdr = TranslateAddress(virtAddress, pageDir, hFile1, hFile2);
	HANDLE hOpen = ChooseFile(&hFile1, &hFile2, &physAdr);
	
	PVOID *buf = new PVOID;
	const UINT POINTSIZE = 8;
	UINT length = len;
	ULONG64 offset = 0;
	CHAR charArray[POINTSIZE*2];

	for(UINT i = 0; i < length; i++)
	{
		if(ReadPhysical(physAdr + offset, buf, POINTSIZE, hOpen) && ReadPhysical(physAdr + offset, charArray, POINTSIZE*2, hOpen))
		{
			_tprintf(_T("%p:  %p "), virtAddress + offset, (ULONG64) *buf);
			offset += POINTSIZE;

			if(ReadPhysical(physAdr + offset, buf, POINTSIZE, hOpen))
			{
				_tprintf(_T("%p "), (ULONG64) *buf);

				for(UINT j = 0; j < POINTSIZE*2; j++)
				{
					CHAR currChar = charArray[j];
					if(ValidAsciiChar(currChar))
					{
						printf(" %c", charArray[j] );
					}
					else
					{
						printf(" .");
					}
				}
				_tprintf(_T("\n"));
			}
			offset += POINTSIZE;
		}
		else
		{
			_tprintf(_T("Error reading physical memory \n"));
		}
	}
	delete[] buf;
	return true;
}

BOOL OutputPhysical(ULONG64 physAddress, HANDLE hFile1, HANDLE hFile2, UINT len)
{
	HANDLE hOpen = ChooseFile(&hFile1, &hFile2, &physAddress);

	PVOID *buf = new PVOID;
	const UINT POINTSIZE = 8;
	UINT length = len;
	ULONG64 offset = 0;
	CHAR charArray[POINTSIZE*2];

	for(UINT i = 0; i < length; i++)
	{
		if(ReadPhysical(physAddress + offset, buf, POINTSIZE, hOpen) && ReadPhysical(physAddress + offset, charArray, POINTSIZE*2, hOpen))
		{
			_tprintf(_T("%p:  %p "), physAddress + offset, (ULONG64) *buf);
			offset += POINTSIZE;

			if(ReadPhysical(physAddress + offset, buf, POINTSIZE, hOpen))
			{
				_tprintf(_T("%p "), (ULONG64) *buf);

				for(UINT j = 0; j < POINTSIZE*2; j++)
				{
					CHAR currChar = charArray[j];
					if(ValidAsciiChar(currChar))
					{
						printf(" %c", charArray[j] );
					}
					else
					{
						printf(" .");
					}
				}
				_tprintf(_T("\n"));
			}
			offset += POINTSIZE;
		}
		else
		{
			_tprintf(_T("Error reading physical memory \n"));
		}
	}
	delete[] buf;
	return true;

}

UINT SearchFile(HANDLE hFile, BYTE *pattern, UINT patLength, UINT noMb, ULONG64 *offsets, UINT maxFindings)
{
	const DWORD DWPAGESIZE = 4096;
	UINT iterations = (1024*1024)*noMb / DWPAGESIZE;
	UINT *ReadBytes = new UINT;
	ULONG64 totalOffset = 0;
	UINT totalMatches = 0;
	ULONG64 matchOffsets[DWPAGESIZE];
	BYTE buffer[DWPAGESIZE];
	BOOL patFound = false;

	for(UINT i = 0; i < iterations; i++)
	{
		//Read I/0 synchronously:
		if(SafeRead(hFile, buffer, DWPAGESIZE, ReadBytes))
		{
			//ULONG64 offsets[DWPAGESIZE];
			UINT matches = 0;
				
			if( matches = Match(buffer, pattern, matchOffsets, *ReadBytes, patLength))
			{
				for(UINT i = 0; i < matches; i++)
				{
					offsets[totalMatches] = matchOffsets[i] + totalOffset;
					totalMatches++;
					if(totalMatches == maxFindings-1)
					{
						_tprintf(_T("Max number of hits found \n"));
						break;

					}
					//_tprintf( _T("Pattern found at offset: %p \n"), matchOffsets[i] + totalOffset);
					patFound = true;
				}
			}
			else
			{
				//_tprintf( _T("Pattern not found. \n"));
			}
			totalOffset += *ReadBytes;
		}
	}
	//Cleanup:
	ReturnFileHandleToStart(hFile);
	delete[] ReadBytes;
	return totalMatches;

}

/*
BOOL TraverseVadTree(ULONG64 vadRootAdr, HANDLE hFile1, HANDLE hFile2, ULONG64 pageDir)
{
	BOOL success = false;
	std::stack<ULONG64> s;
	BALNODE *currNode = rootNode;
	s.push(rootNode->adr() );
	UINT totalsize = 0;
	while( !s.empty() )
	{

		currNode->setadr(s.top(), pDebugClient);
		s.pop();

		//pDebugControl->Output(DEBUG_OUTPUT_NORMAL, "Lock tracker: %p, Length %u bytes \n", currNode->adr(), currNode->length() );
		//MDL *md = new MDL();
		totalsize += currNode->length();

		if( currNode->right() > 0 )
		{
			s.push(currNode->right() );
		}
		if( currNode->left() > 0 )
		{
			s.push(currNode->left() );
		}

	}
	delete currNode;
	return success;
}*/

BOOL TraverseVadTree(ULONG64 vadRootAdr, HANDLE hFile1, HANDLE hFile2, ULONG64 pageDir)
{
	BOOL success = false;
	VAD *currVad = new VAD(hFile1, hFile2, pageDir, vadRootAdr);
	std::stack<ULONG64> s;
	s.push(vadRootAdr);
	_tprintf(_T("Vad:                      Start:            End:      Commit:\n"));
	while(!s.empty())
	{
		currVad->ReadData(hFile1, hFile2, pageDir);
		currVad->PrintData();
		currVad->ReadVadNode(hFile1, hFile2, pageDir, s.top());
		s.pop();

		if(currVad->getRightVad() > 0)
		{
			s.push(currVad->getRightVad());
		}
		if(currVad->getLeftVad() > 0)
		{
			s.push(currVad->getLeftVad());
		}
	}
	delete currVad;
	return success;
}
