#include "VAD.h"
#include "MemoryFunc.h"
#include "MiscFunc.h"

VAD::VAD()
{
	vadAdr = 0;
	vadNode = 0;
	valid = false;
	dataRead = false;
}

VAD::VAD(HANDLE hFile1, HANDLE hFile2, ULONG64 pageDir, ULONG64 vadA)
{
	vadAdr = vadA;
	if(ReadVadNode(hFile1, hFile2, pageDir, vadA))
	{
	}
	else
	{
		vadNode = 0;
		valid = false;
	}
	dataRead = false;
}

VAD::~VAD()
{
	delete vadNode;
}

BOOL VAD::ReadData(HANDLE hFile1, HANDLE hFile2, ULONG64 pageDir)
{
	BOOL success = false;
	if(vadAdr > 0)
	{
		ULONG64 dataAdr = vadAdr + 0x18;
		if(ReadVirtual(dataAdr, pageDir, hFile1, hFile2, rawData, VADSIZE))
		{
			success = true;
			dataRead = true;
		}
		else
		{
			valid = false;
		}
	}
	return success;
}

BOOL VAD::ReadVadNode(HANDLE hFile1, HANDLE hFile2, ULONG64 pageDir, ULONG64 startAdr)
{
	BOOL success = false;
	vadAdr = startAdr;
	//Left = 0x0, Right = 0x8, Parent 0x10
	ULONG64 *leftBuf = new ULONG64;
	ULONG64 *rightBuf = new ULONG64;
	ULONG64 *parentBuf = new ULONG64;
	if(ReadVirtual(startAdr, pageDir, hFile1, hFile2, leftBuf, 8))
	{
		if(ReadVirtual(startAdr+8, pageDir, hFile1, hFile2, rightBuf, 8))
		{
			if(ReadVirtual(startAdr+16, pageDir, hFile1, hFile2, parentBuf, 8))
			{
				vadNode = new BALNODE(*leftBuf, *rightBuf, *parentBuf);
				valid = true;
				success = true;
			}
		}
	}
	delete [] leftBuf;
	delete [] rightBuf;
	delete [] parentBuf;
	return success;
}

VOID VAD::PrintInfo()
{
	if(valid)
	{
		_tprintf(_T("Vad node info: \n"));
		_tprintf(_T("Left Node: %p \n"), vadNode->getLeft());
		_tprintf(_T("Right Node: %p \n"), vadNode->getRight());
		_tprintf(_T("Parent Node: %p \n"), vadNode->getParent());
		_tprintf(_T("\n"));
	}
}

VOID VAD::PrintData()
{
	if(valid && dataRead)
	{
		_tprintf(_T("%p"),vadAdr);
		//Starting virtual address:
		UINT startAdr = ReadUint(rawData, 0);
		BYTE startHigh = rawData[8];
		_tprintf(_T("%*6x%x"),  1, startHigh, startAdr);
		//Ending virtual address:
		UINT endAdr = ReadUint(rawData, 4);
		BYTE endHigh = rawData[9];
		_tprintf(_T("%*6x%x"), 1, endHigh, endAdr);
		/*Commit charge:
		UINT commit = ReadUint(rawData, 28);
		INT sCommit = (INT) commit;
		_tprintf(_T("%4d"), commit);*/
		_tprintf(_T("\n"));
	}
}